from __future__ import absolute_import, division, print_function

__all__ = [
    "__title__", "__summary__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__", "__url__"
]

__title__ = "umbral"

__url__ = "https://github.com/nucypher/pyUmbral"

__summary__ = 'NuCypher\'s Umbral Proxy Re-Encryption Implementation'

__version__ = "0.2.0"

__author__ = "NuCypher"

__email__ = "dev@nucypher.com"

__license__ = "GNU General Public License, Version 3"

__copyright__ = 'Copyright (C) 2019 NuCypher'
